/*
 *  typespeed - measures your typing speed
 *  Copyright (C) 1999-2003   Jani Ollikainen  <bestis@iki.fi>
 *                          & Jaakko Manelius  <jman@iki.fi>
 *  Copyright (C) 2006   Tobias Stoeckmann  <tobias@bugol.de>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * file.c - file access functions
 */

#include <sys/param.h>
#include <sys/types.h>

#include <ctype.h>
#include <curses.h>
#include <dirent.h>
#include <err.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "pathnames.h"
#include "typespeed.h"

#define WORDFILE_MAX	100

struct list {
	int fidelity;
	int pong;
	float cps;
	float tcps;
	float typoratio;
	char faim[21];
};

extern void	 drawscreen(void);
extern int	 level(int);
extern void	 liima_mvgetnstr(int, int, char *, int);
extern void	 pressanykey(int, int);
extern int	 typorankkaus(float);
extern void	*xmalloc(size_t);
extern void	 xcolor_set(short);

int	addtop10(char *, struct stats *);
int	choose_wordfile(int);
void	freewords(void);
int	iswordlist(struct dirent *);
void	loadhighscores(char *);
int	loadwords(char *);
void	makescorefiles(void);
void	readconfig(void);
int	resetscorefile(char *);

extern char	*rankki[11];
extern char	*typorank[12];
extern char	*word[WORD_MAX];
extern char	*usedwordfile;
extern int	 wordcount;

/*
 * Reads in max WORD_MAX words out of file "filename" into word. If
 * less than 22 words are parsed, all words are freed and 1 is returned.
 * Words cannot be longer than 19 chars. No word duplicates allowed.
 *
 * First line of file "filename" will be ignored (description).
 *
 * XXX: loadwords is able to add words to an already populated
 * word array, I think this is not needed anywhere, so this
 * functionality will be removed.
 *
 * Returns 0 on success or 1 on error.
 */
int
loadwords(char *filename)
{
	FILE *wordlist;
	char *myword[WORD_MAX];
	char *p;
	char newword[19 + 3], wordpath[MAXPATHLEN];
	int ignore, j, n;
	size_t i, mycount;
	ignore = 1;

	n = snprintf(wordpath, sizeof(wordpath), "%s/%s", worddir, filename);
	if (n < 1 || (size_t)n > sizeof(wordpath) - 1)
		return 1;

	if ((wordlist = fopen(wordpath, "r")) == NULL)
		return 1;

	memset(myword, 0, sizeof(myword));
	mycount = 0;
	while (fgets(newword, sizeof(newword), wordlist) != NULL) {
		/*
		 * we won't freak out about words which are against
		 * the rule (e.g. too long). Instead we will silently
		 * ignore them.
		 */
		if ((p = strchr(newword, '\n')) == NULL) {
			ignore = 1;
			continue;
		}
		if (ignore) {
			ignore = 0;
			continue;
		}

		*p = '\0';
		if ((p = strchr(newword, '\r')) != NULL)
			*p = '\0';

		if (strlen(newword) > 19 || !strlen(newword))
			continue;

		for (ignore = 0, p = newword; *p != '\0'; p++)
			if (!isprint(*p) || isspace(*p)) {
				ignore = 1;
				break;
			}
		if (ignore) {
			ignore = 0;
			continue;
		}

		if (mycount >= WORD_MAX)
			break;

		myword[mycount] = xmalloc(strlen(newword) + 1);
		strcpy(myword[mycount++], newword);
	}
	fclose(wordlist);

	for (i = 0; i < mycount; i++) {
		for (j = 0; j < WORD_MAX; j++) {
			if (word[j] == NULL) {
				if (wordcount < WORD_MAX) {
					wordcount++;
					word[j] = myword[i];
				}
				break;
			}
			if (!strcmp(myword[i], word[j])) {
				free(myword[i]);
				break;
			}
		}
		if (wordcount == WORD_MAX)
			break;
	}

	if (wordcount < 22) {
		/* not enough words */
		freewords();
		return 1;
	}
	return 0;
}

/* Frees all allocated memory in word. */
void
freewords(void)
{
	while (wordcount) {
		wordcount--;
		free(word[wordcount]);
		word[wordcount] = NULL;
	}
}

/*
 * A huge function that also uses curses to print a menu.
 * In this menu you can select found wordlists for further
 * processing. The processing of the wordlist depends on
 * "operation":
 *
 * 0: loads words out of wordlist
 * 1: prints current highscore for wordlist
 * 2: asks if specified highscore should be erased (reset)
 *
 * Returns 0 on success or 1 on failure.
 */
int
choose_wordfile(int operation)
{
	FILE *fileinfo;
	struct dirent **namelist;
	char *p;
	char info[61], thingie[WORDFILE_MAX][61];
	char wordpath[MAXPATHLEN];
	int a, ch, cpos, k, n, retval, tot, w;
	int *what;

	drawscreen();
	xcolor_set(5);
	if ((n = scandir(worddir, &namelist, iswordlist, alphasort)) < 0)
		err(1, "scandir()");
	if (!n) {
		mvaddstr(1, 5, "No wordlists found...");
		pressanykey(3, 5);
		return 1;
	}

	if ((what = calloc(n, sizeof(int))) == NULL)
		err(1, NULL);

	k = 0;
	while (n--) {
		w = snprintf(wordpath, sizeof(wordpath), "%s/%s", worddir, namelist[n]->d_name);
		if (w < 1 || (size_t)w > sizeof(wordpath) - 1)
			continue;

		if ((fileinfo = fopen(wordpath, "r")) != NULL) {
			if (fgets(info, 61, fileinfo) != NULL) {
				if ((p = strchr(info, '\n')) != NULL)
					*p = '\0';
				if ((p = strchr(info, '\r')) != NULL)
					*p = '\0';
				strcpy(thingie[k], info);
				what[k++] = n;
			}
			fclose(fileinfo);
		}
	}

	tot = k;
	xcolor_set(4);
	mvaddstr(2, 5, "Choose a wordfile (UP/DOWN/ENTER):");
	if (tot > 17)
		for (k = 0; k < 17; k++)
			mvaddstr(k + 4, 5, thingie[k]);
	else
		for (k = 0; k < tot; k++)
			mvaddstr(k + 4, 5, thingie[k]);

	cpos = k = 0;
	mvaddstr(4, 1, "->");

	a = 1;
	while (a) {
		/*
		 * and thank's for xmunkki, he made my day and
		 * "solved" this, what i didn't get right (or did
		 * but i used 2 extra variables.. :)
		 */
		while ((ch = getch()) == ERR);
		switch (ch) {
		case 'A':
			if (cpos > 0) {
				cpos--;
				if (k > 0)
					k--;
			} else if (k > 0) {
				k--;
				move(20, 5);
				deleteln();
				move(4, 5);
				insertln();
				mvaddstr(4, 5, thingie[k]);
			}
			break;
		case 'B':
			if ((cpos < 16) && (cpos < tot - 1)) {
				cpos++;
				if (k < tot - 1)
					k++;
			} else if (k < tot - 1) {
				k++;
				move(4, 5);
				deleteln();
				move(20, 5);
				insertln();
				mvaddstr(20, 5, thingie[k]);
			}
			break;
		case ' ':
		case '\n':
			a = 0;
			break;
		default:
			break;
		}
		mvaddstr(cpos + 3, 1, "  ");
		mvaddstr(cpos + 5, 1, "  ");
		mvaddstr(cpos + 4, 1, "->");
	}

	a = 1;
	switch (operation) {
	case 2:
		loadhighscores(namelist[what[k]]->d_name);
		xcolor_set(4);
		mvaddstr(16, 3, "Do you want to RESET this [y/n]? ");
		do {
			while ((ch = getch()) == ERR);
			ch = tolower(ch);
		} while ((ch != 'y') && (ch != 'n'));
		a = 0;
		if (ch == 'y')
			a = resetscorefile(namelist[what[k]]->d_name);
		else
			break;
		/* FALLTHROUGH */
	case 1:
		loadhighscores(namelist[what[k]]->d_name);
		pressanykey(16, 3);
		a = 0;
		break;
	case 0:
		a = loadwords(namelist[what[k]]->d_name);
		break;
	default:
		break;
	}

	if (a) {
		xcolor_set(2);
		mvaddstr(3, 5, "Something went wrong. Contact your mother.");
		while (getch() == ERR);
		retval = 1;
	}
	else {
		if ((usedwordfile = strdup(namelist[what[k]]->d_name)) == NULL)
			err(1, "strdup()");
		retval = 0;
	}

	for (k = 0; k < tot; k++)
		free(namelist[k]);
	free(namelist);
	free(what);
	return retval;
}

/*
 * This is a helper function to convert struct stats to struct list.
 * Remember to free the return value later on!
 */
struct list *
newentry(int epong, int efidelity, float ecps, float etcps, float etyporatio, char efaim[21])
{
	struct list *entry;

	if ((entry = calloc(1, sizeof(struct list))) == NULL)
		err(1, NULL);

	entry->pong = epong;
	entry->fidelity = efidelity;
	entry->cps = ecps;
	entry->tcps = etcps;
	entry->typoratio = etyporatio;
	strncpy(entry->faim, efaim, sizeof(entry->faim));
	entry->faim[20] = '\0';

	return entry;
}

/*
 * Writes default values into highscore file for "sanafile".
 * Returns 0 on success or 1 on failure.
 */
int
resetscorefile(char *sanafile)
{
	FILE *highfile;
	int i, n, rvalue;
	char filetys[MAXPATHLEN];
	struct list *entry[10];

	entry[0] = newentry(700, 8, 7  , 7.5, 10, "3l33t");
	entry[1] = newentry(600, 7, 6  , 6.5, 10, "Pro");
	entry[2] = newentry(500, 6, 5  , 5.5, 10, "VeryGood");
	entry[3] = newentry(400, 5, 4  , 4.5, 10, "Good");
	entry[4] = newentry(300, 4, 3  , 3.5, 10, "Average");
	entry[5] = newentry(200, 3, 2  , 2.5, 10, "NoGood");
	entry[6] = newentry(100, 2, 1  , 1.5, 10, "ADP-Pro");
	entry[7] = newentry( 50, 1, 0.5, 1  , 10, "Loser");
	entry[8] = newentry(  0, 0, 0  , 0  ,  0, "None");
	entry[9] = newentry(  0, 0, 0  , 0  ,  0, "None");

	n = snprintf(filetys, sizeof(filetys), "%s/high.%s", highdir, sanafile);
	if (n < 6 || (size_t)n > sizeof(filetys) - 1)
		return 1;

	if ((highfile = fopen(filetys, "w")) != NULL) {
		for (i = 0; i < 10; i++) {
			fwrite(entry[i], sizeof(struct list), 1, highfile);
			free(entry[i]);
		}
		fclose(highfile);
		rvalue = 0;
	} else
		rvalue = 1;

	return rvalue;
}

/*
 * Loads highscore for "sanafile", where "sanafile" contains the name
 * of the wordlist and gives out a highscore list.
 */
void
loadhighscores(char *sanafile)
{
	FILE *highfile;
	int n, x;
	char *p;
	char filetys[MAXPATHLEN];
	struct list high[10];

	n = snprintf(filetys, sizeof(filetys), "%s/high.%s", highdir, sanafile);
	if (n < 6 || (size_t)n > sizeof(filetys) - 1)
		return;

	drawscreen();
	xcolor_set(4);

	if ((highfile = fopen(filetys, "r")) != NULL) {
		mvaddstr(1, 0,
			 "Top10  (score) (name)                   (level)   (cps)  (tcps) (typoinfos)");
		xcolor_set(5);

		if (fread(&high, sizeof(struct list), 10, highfile) != 10) {
			mvaddstr(4, 3, "Corrupt Highscore file");
			return;
		}
		fclose(highfile);

		for (x = 0; x < 10; x++) {
			xcolor_set(5);
			mvprintw(x + 2, 2, "%2d.", x + 1);
			mvprintw(x + 2, 7, "%6d", high[x].pong);
			high[x].faim[20] = '\0';
			for (p = high[x].faim; *p != '\0'; p++)
				if(iscntrl(*p)) {
					strcpy(high[x].faim, "Kiddie");
					break;
				}
			mvaddstr(x + 2, 15, high[x].faim);
			level(high[x].pong);
			if (high[x].fidelity >= 0 && high[x].fidelity < 11)
				mvaddstr(x + 2, 40, rankki[high[x].fidelity]);
			xcolor_set(5);
			now.level = 0;	/* Why here?? */
			mvprintw(x + 2, 50, "%2.3f", high[x].cps);
			mvprintw(x + 2, 57, "%2.3f", high[x].tcps);
			mvprintw(x + 2, 64, "%2.2f%%", high[x].typoratio);
			if (high[x].pong) {
				/* typorankkaus(high.typoratio); */
				mvaddstr(x + 2, 71, typorank[typorankkaus(high[x].typoratio)]);
			} else
				mvaddstr(x + 2, 71, "None");
		}
	} else
		mvaddstr(4, 3, "No Highscores");
}

/*
 * Adds a new entry into highscore for "sanaf", where "sanaf"
 * contains name of wordlist. "stat" will be converted into
 * highscore fileformat and written into highscore file. Should this
 * file be empty, create a new standard highscore before adding new
 * entry.
 *
 * A note will be written to screen.
 *
 * Returns 0 on success or 1 on failure.
 */
int
addtop10(char *sanaf, struct stats *stat)
{
	FILE *highf;
	char filet[MAXPATHLEN];
	char name[21];
	int madeit, n, x;
	struct list *entry;
	struct list high[10];

	n = snprintf(filet, sizeof(filet), "%s/high.%s", highdir, sanaf);
	if (n < 6 || (size_t)n > sizeof(filet) - 1)
		return 1;

	if ((highf = fopen(filet, "r")) != NULL)
		fclose(highf);
	else
		resetscorefile(sanaf);

	if ((highf = fopen(filet, "r")) != NULL) {
		if (fread(&high, sizeof(struct list), 10, highf) != 10) {
			return 1;
		}
		fclose(highf);
	}

	if ((highf = fopen(filet, "w")) != NULL) {
		madeit = 0;
		for (x = 0; x < 10; x++) {
			if (stat->score > high[x].pong && !madeit) {

				drawscreen();

				xcolor_set(5);
				mvaddstr(3, 2, "You managed to get to top10!");
				mvaddstr(5, 2, "Enter your name:");
				liima_mvgetnstr(5, 19, name, 20);

				entry = newentry(stat->score, stat->level, stat->speed, stat->totalspeed, stat->ratio, name);
				fwrite(entry, sizeof(struct list), 1, highf);
				free(entry);

				madeit = 1;

				fwrite(&high[x], sizeof(struct list), 1, highf);
			}
			else
				if ((madeit == 1) && (x == 9)) {
                                 } else
					fwrite(&high[x], sizeof(high[x]), 1, highf);
		}
		fclose(highf);
	}

	return 0;
}

/*
 * Parses config files:
 * - global configuration
 * - user-specific configuration
 *
 * Both files should not contain more than 2 lines, the first line
 * contains the directory of wordlists, the second line could contain
 * the word "cheat" which would activate the cheat mode.
 *
 * User-specific configuration overwrites global config, but a
 * global cheat mode cannot be overwritten with another config file.
 */
void
readconfig(void)
{
	FILE *conffile;
	char wordhakemisto[MAXPATHLEN + 3];
	char tmp[1024];
	char userhome[MAXPATHLEN];
	char *p, *envhome;
	int n;

	n = snprintf(highdir, sizeof(highdir), "%s", HIGHDIR);
	if (n < 0 || (size_t)n > sizeof(highdir) - 1) {
		highdir[0] = '.';
		highdir[1] = '\0';
	}
	n = snprintf(worddir, sizeof(worddir), "%s", WORDDIR);
	if (n < 0 || (size_t)n > sizeof(worddir) - 1) {
		worddir[0] = '.';
		worddir[1] = '\0';
	}

	if ((conffile = fopen(CONFIGFILE, "r")) != NULL) {
		/* read in wordlist directory */
		if (fgets(wordhakemisto, sizeof(wordhakemisto), conffile) != NULL) {
			if ((p = strchr(wordhakemisto, '\n')) == NULL) {
				while (fgets(tmp, sizeof(tmp), conffile) != NULL) {
					if (strchr(tmp, '\n') != NULL)
						break;
				}
			}
			else {
				*p = '\0';
				if ((p = strchr(wordhakemisto, '\r')) != NULL)
					*p = '\0';
				if (strlen(wordhakemisto) < sizeof(worddir) && wordhakemisto[0] == '/')
					strncpy(worddir, wordhakemisto, sizeof(worddir));
			}
		}
		/* activate cheat? */
		if ((fgets(tmp, sizeof(tmp), conffile)) != NULL) {
			if ((p = strchr(tmp, '\n')) != NULL)
				*p = '\0';
			if ((p = strchr(tmp, '\r')) != NULL)
				*p = '\0';
			if (!strncmp(tmp, "cheat", strlen("cheat") + 1))
				opt.cheat = 1;
		}
		fclose(conffile);
	}

	if ((envhome = getenv("HOME")) == NULL)
		return;

	n = snprintf(userhome, sizeof(userhome), "%s/%s", envhome, LOCALCONF);
	if (n < 1 || (size_t)n > sizeof(userhome) - 1)
		return;

	if ((conffile = fopen(userhome, "r")) != NULL) {
		/* read in wordlist directory */
		if (fgets(wordhakemisto, sizeof(wordhakemisto), conffile) != NULL) {
			if ((p = strchr(wordhakemisto, '\n')) == NULL) {
				while (fgets(tmp, sizeof(tmp), conffile) != NULL) {
					if (strchr(tmp, '\n') != NULL)
						break;
				}
			}
			else {
				*p = '\0';
				if ((p = strchr(wordhakemisto, '\r')) != NULL)
					*p = '\0';
				if (strlen(wordhakemisto) < sizeof(worddir) && wordhakemisto[0] == '/')
					strncpy(worddir, wordhakemisto, sizeof(worddir));
			}
		}
		/* activate cheat? */
		if ((fgets(tmp, sizeof(tmp), conffile)) != NULL) {
			if ((p = strchr(tmp, '\n')) != NULL)
				*p = '\0';
			if ((p = strchr(tmp, '\r')) != NULL)
				*p = '\0';
			if (!strncmp(tmp, "cheat", strlen("cheat") + 1))
				opt.cheat = 1;
		}
		fclose(conffile);
	}
}

/* Creates default highscorefiles for each wordlist. */
void
makescorefiles(void)
{
	int n;
	struct dirent **namelist;

	if ((n = scandir(worddir, &namelist, iswordlist, alphasort)) < 0)
		err(1, "scandir()");
	if (n) {
		puts("\nResetting/Making scorefiles for all found wordfiles...\n");
		while (n--) {
			fputs("Make scorefile for wordfile: ", stdout);
			puts(namelist[n]->d_name);
			resetscorefile(namelist[n]->d_name);
			free(namelist[n]);
		}
		free(namelist);
	}
}

/*
 * Filter function for scandir: we just want words.*-files.
 * Returns 1 if file is expected to be a wordlist or 0 if it's not.
 */
int
iswordlist(struct dirent *entry)
{
	if (!strncmp(entry->d_name, "words.", strlen("words.")))
		return 1;
	return 0;
}

