/*
 *  typespeed - measures your typing speed
 *  Copyright (C) 1999-2003   Jani Ollikainen  <bestis@iki.fi>
 *                          & Jaakko Manelius  <jman@iki.fi>
 *  Copyright (C) 2006   Tobias Stoeckmann  <tobias@bugol.de>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * typespeed.c - the main code
 */

#include <sys/types.h>
#include <sys/socket.h>

#include <ctype.h>
#include <curses.h>
#include <errno.h>
#include <getopt.h>
#include <locale.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>

#include "typespeed.h"
#include "version.h"

#define DEFAULT_CHEAT		0
#define DEFAULT_CLIENT		0
#define DEFAULT_COLORS		1
#define DEFAULT_PORT		6025
#define DEFAULT_SERVER		0
#define DEFAULT_TRAINING	0
#define DEFAULT_TRAININGSPEED	2

extern int	 addtop10(char *, struct stats *);
extern int	 choose_wordfile(int);
extern void	 drawmenu(void);
extern void	 drawstatus(int);
extern void	 drawscreen(void);
extern void	 endcursestuff(void);
extern void	 freewords(void);
extern int	 init_network(char *, int);
extern void	 initcursestuff(void);
extern int	 level(int);
extern void	 loadhighscores(char *);
extern void	 makescorefiles(void);
extern void	 multipmenu(void);
extern void	 net_swapscore(int, struct stats *, struct stats *);
extern void	 net_waitgame(int);
extern void	 optionmenu(void);
extern void	 pressanykey(int, int);
extern int	 r(int);
extern void	 readconfig(void);
extern void	 showhighscores(void);
extern void	 tell_story(void);
extern clock_t	 timenow(void);
extern int	 typorankkaus(float);
extern void	*xmalloc(size_t);
extern void	 xcolor_set(short);

/* globals */
char *rankki[11] =
{"None", "Loser", "ADP-Pro", "NoGood", "Average",
 "Good", "VeryGood", "Pro", "3l33t", "*(GOD)*", "Computer"};

char *typorank[12] =
{"XMunkki", "Alien", "Secretary", "Human", "IT Person", "Handicap",
 "Monkey", "Pencil", "T-Bone", "E-Typo", "TypOmatiC", "TypoKing"};

struct stats	 best;
int		 misses;
struct stats	 other;
float		 rate;
int		 ttss;
int		 typesocket = 0;
char		*usedwordfile;
char		*word[WORD_MAX];
int		 wordcount = 0;
int		 wordpos[22];
char		 wordstring[22][20];

struct option options[] = {
	{"nocolors", no_argument, &opt.usecolors, 0},
	{"cheat", no_argument, &opt.cheat, 1},
	{"training", optional_argument, NULL, 't'},
	{"client", required_argument, NULL, 'o'},
	{"server", no_argument, &opt.server, 1},
	{"port", required_argument, NULL, 'p'},
	{"makescores", no_argument, NULL, 'm'},
	{"help", no_argument, NULL, 'h'},
	{0, 0, 0, 0}
};

/* Writes "length" blanks at position x y. */
void
clearword(int y, int x, size_t length)
{
	char *blankword;
	blankword = xmalloc(length + 1);
	memset(blankword, ' ', length);
	blankword[length] = '\0';
	mvaddstr(y, x, blankword);
	free(blankword);
}

/*
 * Move words across the screen.
 * Removes them if they reach the right side and increases misses.
 */
void
movewords(void)
{
	int i;
	int length;

	for (i = 0; i < 22; i++) {
		length = (int)strlen(wordstring[i]);
		if (wordpos[i] > (65 - length))
			xcolor_set(3);
		else if (wordpos[i] > (50 - length))
			xcolor_set(7);
		else
			xcolor_set(1);

		if (wordpos[i] > (79 - length)) {
			clearword(i, wordpos[i], length);
			misses++;
			wordpos[i] = 0;
		} else if (wordpos[i] > 0)
			mvprintw(i, wordpos[i]++, " %s", wordstring[i]);
	}
}

/*
 * Insert a new word into wordstring and wordpos, i.e. one more word
 * on the screen.
 *
 * If newword points to a string, newword will be inserted.
 * If newword is NULL, a random word out of word will be inserted.
 * In both cases, duplicates are not allowed. Should newword point
 * to duplicated string, no string will be inserted.
 *
 * Returns TRUE on success or FALSE on failure.
 *
 * XXX: Can result in a veeery long loop when there are many words on
 *      field and only a few words in wordlist (no duplicates allowed).
 */
int
addnewword(char *newword)
{
	char *myword;
	int count, dup, slot;
	int freeslot[22];
	size_t i, length;

	/* take a random free slot */
	for (count = 0, i = 0; i < 22; i++) {
		if (!wordpos[i])
			freeslot[count++] = i;
	}
	if (!count) {
		/* the player has enough trouble... */
		return FALSE;
	}
	slot = freeslot[r(count)];

	if ((myword = newword) == NULL)
		myword = word[r(wordcount)];

	do {
		for (dup = 0, i = 0; i < 22; i++)
			if (!strcmp(wordstring[i], myword)) {
				if (newword != NULL)
					return FALSE;
				dup = 1;
				myword = word[r(wordcount)];
				break;
			}
	} while (dup);

	if ((length = strlen(myword)) > 19)
		return FALSE;

	/* lower string */
	for (i = 0; i < length; i++)
		myword[i] = tolower(myword[i]);

	strncpy(wordstring[slot], myword, sizeof(wordstring[slot]));
	wordpos[slot] = 1;
	return TRUE;
}

/*
 * Sets rate according to current length of all visible words.
 * If words with too few chars are on screen, throw in another word.
 */
void
cwords(void)
{
	int i, length;

	for (length = 0, i = 0; i < 22; i++)
		if (wordpos[i] > 0)
			length += strlen(wordstring[i]) + 1;
	if (length < now.score / 4 + 1) {
		addnewword(NULL);
		rate = now.score / 175 + 3;
	}
}

int
play(void)
{
	char buf[256];
	char input[20];

	clock_t oldtimes, starttime;
	clock_t pausetime = 0;

	int inputpos;
	int i;
	int nappaykset;
	int wordswritten = 0;
	int tra = 0;
	float min10;
	int nappi, escend;
	srandom((unsigned)time(NULL));

	if ((opt.client || opt.server) && opt.training)
		opt.training = 0;

	if ((i = choose_wordfile(0)) == 1)
		return 1;
	drawscreen();
	nodelay(stdscr, 1);
	cbreak();

	xcolor_set(2);

	if (opt.server || opt.client) {
		mvaddstr(12, 20, "Waiting for the other party to join in...");
		while (getch() != ERR);
		net_waitgame(typesocket);
		flushinp();
		drawscreen();
	}

	misses = 0;
	min10 = 0;
	nappaykset = 0;
	now.score = 0;
	inputpos = 0;
	escend = 0;
	rate = 1;
	now.ratio = 0;

	memset(wordpos, 0, sizeof(wordpos));

	halfdelay(1);
	oldtimes = timenow();

	starttime = timenow();
	input[0] = '\0';

	while (misses < 10 && escend < 1) {
		while (timenow() < (oldtimes + (100.0 / rate))) {
			if ((nappi = getch()) == ERR)
				continue;

			switch (tolower(nappi)) {
			case 8: /* BACKSPACE */
			case 127: /* DEL */
				if (inputpos > 0) {
					mvaddch(23, 1 + inputpos, ' ');
					input[--inputpos] = '\0';
				}
				break;
			case 27: /* ESC (sequence) */
				escend = 1;
				if (opt.client || opt.server)
					break;
				for (i = 0; i < 3; i++)
					/* CURSOR UP is legal */
					if (getch() == 65)
						escend = 0;
				if (escend)
					break;
				pausetime = timenow();
				mvaddstr(23, 2, "       PAUSED      ");
				move(23, 2);
				while (getch() == ERR);
				starttime += timenow() - pausetime;
				mvaddstr(23, 9, "      ");
				flushinp();
				/* FALLTHROUGH */
			case 21: /* ^U */
				if (inputpos > 0) {
					clearword(23, 2, inputpos);
					inputpos = 0;
					input[0] = '\0';
				}
				break;
			case 32: /* SPACE */
			case 10: /* ENTER */
				nappaykset++;
				for (i = 0; i < 22; i++)
					if (!strcmp(input, wordstring[i])) {
						now.score += inputpos;
						wordswritten++;
						clearword(i, wordpos[i], strlen(wordstring[i]));
						wordpos[i] = 0;
						/* send 50% words in multiplayer */
						if (opt.server || opt.client) {
							sprintf(buf, "%s\n", wordstring[i]);
							if (!r(2))
								send(typesocket, buf, strlen(buf), 0);
						}
						strcpy(wordstring[i], "THIS-IS-NOT-A-CHEAT");
					}
				clearword(23, 2, 19);
				inputpos = 0;
				input[0] = '\0';
				break;
			case 11: /* ^K */
				if (inputpos < (int)strlen(input)) {
					input[inputpos] = '\0';
					clearword(23, 2 + inputpos, 19 - inputpos);
				}
				break;
			case 6: /* ^F */
				if (inputpos < 19 && inputpos < (int)strlen(input))
					inputpos++;
				break;
			case 5: /* ^E */
				if (inputpos != (int)strlen(input))
					inputpos = strlen(input);
				break;
			case 2: /* ^B */
				if (inputpos > 0)
					inputpos--;
				break;
			case 1: /* ^A */
				if (inputpos > 0)
					inputpos = 0;
				break;
			default:
				nappaykset++;

				if (inputpos < 19 && !iscntrl(nappi)) {
					mvaddch(23, 2 + inputpos, nappi);
					if (inputpos == (int)strlen(input))
						input[inputpos + 1] = '\0';
					input[inputpos++] = nappi;
					if (opt.cheat) {
						for (i = 0; i < 22; i++)
							if (!strcmp(input, wordstring[i])) {
								now.score += inputpos;
								clearword(i, wordpos[i], strlen(wordstring[i]));
								wordpos[i] = 0;
								clearword(23, 2, 19);
								inputpos = 0;
								input[0]= '\0';
								strcpy(wordstring[i], "THIS-IS-NOT-A-CHEAT");
							}
					}
				}
				break;
			}
			move(23, 2 + inputpos);
		}

		oldtimes = timenow();

		if (opt.training) {
			if (tra == opt.traspeed) {
				movewords();
				cwords();
				tra = 0;
			}
			tra++;
		}
		else {
			movewords();
			cwords();
		}

		now.speed = (now.score + wordswritten) * 100.0 / (timenow() - starttime);
		now.totalspeed = nappaykset * 100.0 / (timenow() - starttime);

		drawstatus(inputpos);
		xcolor_set(2);

		/* last, we check if the opponent has any new words for us... */

		if (opt.server || opt.client) {
			i = recv(typesocket, buf, 256, MSG_PEEK);
			buf[255] = '\0';

			if (i == -1 && errno != EWOULDBLOCK) {
				perror("recv()");
				exit(1);
			}
			if (i > 0) {
				char *j = strchr(buf, '\n');
				if (j != NULL) {
					*j = '\0';
					addnewword(buf);
					recv(typesocket, buf, strlen(buf) + 1, 0);
					if (!strncmp(buf, "ENDOFTHEGAME", 12))
						escend = 2;
				}
			}
		}
		refresh();
	}

	if ((opt.server || opt.client) && escend != 2) {
		strcpy(buf, "ENDOFTHEGAME\n");
		send(typesocket, buf, strlen(buf), 0);
	}

	endcursestuff();
	initcursestuff();
	drawscreen();
	now.level = level(now.score);
	if (now.level >= best.level && now.score >= best.score) {
		best.level = now.level;
		best.score = now.score;
		best.speed = now.speed;
		best.totalspeed = now.totalspeed;
		if (now.totalspeed && now.speed)
			best.ratio = (now.totalspeed - now.speed) / now.totalspeed;
	}

	if (opt.server || opt.client) {
		net_swapscore(typesocket, &now, &other);
		while (getch() != ERR);
	}
	xcolor_set(5);

	mvaddstr(12, 30, "GAME OVER!");
	refresh();

	sleep(3);

	/* total speed */
	min10 = now.speed * 600;

	/*
	 * let's get speed away from your mistakes or not?
	 * min10=min10-(ratio*100)
	 */

	mvaddstr(12, 30, "          ");
	mvprintw(3, 20, "Typespeed %s ", TVERSION);

	mvaddstr(5, 20, "Your score was: ");
	if (opt.client||opt.server)
		mvaddstr(5, 60, "Opponent's: ");
	mvaddstr(7, 20, "Rank:\t\t");
	/* level(score); */
	mvaddstr(7, 40, rankki[level(now.score)]);
	xcolor_set(5);
	mvprintw(8, 20, "Score:\t\t%d", now.score);
	mvprintw(9, 20, "10MRS:\t\t%1.0f", min10);
	mvprintw(10, 20, "Total CPS:\t\t%2.3f  ", now.totalspeed);
	mvprintw(11, 20, "Correct CPS:\t%2.3f  ", now.speed);
	if (now.totalspeed && now.speed)
		now.ratio = (now.totalspeed - now.speed) / now.totalspeed * 100;
	mvprintw(12, 20, "Typo ratio:\t\t%2.1f%%\t", now.ratio);
	mvaddstr(13, 20, "Typorank:\t\t");

	if (now.score)
		mvaddstr(13, 40, typorank[typorankkaus(now.ratio)]);
	else
		mvaddstr(13, 40, "None");

	if (opt.server || opt.client) {
		mvaddstr(7, 60, rankki[level(other.score)]);
		xcolor_set(5);
		mvprintw(8, 60, "%d", other.score);
		mvprintw(9, 60, "%1.0f", other.speed * 600);
		mvprintw(10, 60, "%2.3f  ", other.totalspeed);
		mvprintw(11, 60, "%2.3f  ", other.speed);
		if (other.totalspeed && other.speed)
			other.ratio = (other.totalspeed - other.speed) / other.totalspeed * 100;
		mvprintw(12, 60, "%2.1f%%", other.ratio);
		if (other.score)
			mvaddstr(13, 60, typorank[typorankkaus(other.ratio)]);
		else
			mvaddstr(13, 60, "None");
	}

	xcolor_set(2);
	if (opt.training)
		mvaddstr(17, 15, "TRAINING MODE - YOU CANNOT GET TO TOP10");
	else if (opt.client || opt.server)
		mvaddstr(17, 20, "NETWORK MODE - YOU CANNOT GET TO TOP10");
	nodelay(stdscr, 0);
	cbreak();
	flushinp();

	pressanykey(15, 20);
	if (!opt.training && !opt.client && !opt.server)
		addtop10(usedwordfile, &now);
	loadhighscores(usedwordfile);

	pressanykey(16, 3);
	free(usedwordfile);
	freewords();
	return 0;
}

/* Prints usage information on stderr and exits with return value 1. */
void
usage(char *prog)
{
	fputs("\nCommand line options:\n", stderr);
	fputs("  --cheat         == multimedia cheat(?)\n", stderr);
	fputs("  --client=addr   == IP address of server\n", stderr);
	fputs("  --makescores    == Reset/Make new scorefiles\n", stderr);
	fputs("  --nocolors      == don't use colors\n", stderr);
	fputs("  --port=port     == Port number in the network play\n", stderr);
	fputs("  --server        == start typespeed in server mode\n", stderr);
	fputs("  --training[=nm] == training mode\n", stderr);
	fputs("                     (where nm is speed number)\n", stderr);
	fputs("  --help          == this help!\n\n", stderr);
	fprintf(stderr, "Usage: %s options\n", prog);
	exit(1);
}

int
main(int argc, char **argv)
{
	int i, wheretogo;
	char serv[256];

	setlocale(LC_CTYPE, "");

	opt.cheat = DEFAULT_CHEAT;
	opt.training = DEFAULT_TRAINING;
	opt.traspeed = DEFAULT_TRAININGSPEED;
	opt.usecolors = DEFAULT_COLORS;
	opt.server = DEFAULT_SERVER;
	opt.client = DEFAULT_CLIENT;
	opt.port = DEFAULT_PORT;

	memset(&best, 0, sizeof(best));
	memset(&now, 0, sizeof(now));
	memset(&word, 0, sizeof(word));

	readconfig();
	printf("Typespeed %s, Copyright (C) Jani Ollikainen & Jaakko Manelius\n", TVERSION);
	puts("                  Copyright (C) Tobias Stoeckmann");
	puts("Typespeed comes with ABSOLUTELY NO WARRANTY; for details read COPYING.");

	while ((i = getopt_long(argc, argv, "to:h", options, NULL)) != -1) {
		switch (i) {
		case 0:
			/* flag already set by getopt */
			break;
		case 'm':
			makescorefiles();
			exit(0);
			break;
		case 'o':
			opt.client = 1;
			if (!optarg) {
				/* getopt should handle this */
				abort();
			}
			strncpy(serv, optarg, 255);
			serv[255] = '\0';
			break;
		case 'p':
			if (optarg)
				opt.port = atoi(optarg);
			if (opt.port <= 1024) {
				printf("DON'T USE PRIVILEGED PORT\n");
				exit(1);
			}
			break;
		case 't':
			opt.training = 1;
			if (optarg)
				opt.traspeed = atoi(optarg);
			if (opt.traspeed < 2 || opt.traspeed > 10)
				opt.traspeed = 2;
			break;
		case ':':
		case '?':
			/* getopt already printed error */
		case 'h':
		default:
			usage(argv[0]);
			break;
		}
	}

	if ((opt.server || opt.client) && opt.training)
		opt.training = 0;

	if (opt.server || opt.client)
		while ((typesocket = init_network(serv, 0)) < 0);

	sleep(1);

	initcursestuff();

	if (typesocket > 0)
		play();

	drawmenu();
	/*
	 * THIS IS OLD NEWS - LEFT HERE TO PUT SOME FUTURE NEWS IN
	 * mvaddstr(15,20, "NOTE THIS VERSION HAS NEW WAY TO CALCULATE SPEED!");
	 * mvaddstr(16,26, "SCORES ARE NOT THE SAME AS THEY WERE");
	 */

	wheretogo = 0;
	do {
		nodelay(stdscr, 0);
		cbreak();
	
		while ((wheretogo = getch()) == ERR);

		switch (wheretogo) {
		case '1':
			play();
			break;
		case '2':
			if (opt.training) {
				opt.training = 0;
				drawscreen();
			}
			multipmenu();
			if (typesocket > 0)
				play();
			break;
		case '3':
			tell_story();
			break;
		case '4':
			showhighscores();
			break;
		case '5':
			optionmenu();
			break;
		default:
			break;
		}
		if (wheretogo > '0' && wheretogo < '6')
			drawmenu();
	} while (wheretogo != '6');

	endcursestuff();

	printf("\e[H\e[J");
	printf("Typespeed %s\n\n", TVERSION);
	printf("Best score was: \n\n");
	printf("Rank:\t\t%s\n", rankki[best.level]);
	printf("Score:\t\t%d\n", best.score);
	printf("Total CPS:\t%2.3f\n", best.totalspeed);
	printf("Correct CPS:\t%2.3f\n", best.speed);
	printf("Typo ratio:\t%2.1f%%\n", best.ratio * 100);
	printf("Typorank:\t");
	if (best.score)
		puts(typorank[typorankkaus(best.ratio * 100)]);
	else
		puts("None");
	puts("\n");
	return 0;
}

